/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.registry.core.manager;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.petals.registry.api.Information;
import org.ow2.petals.registry.api.LocalRegistry;
import org.ow2.petals.registry.api.Query;
import org.ow2.petals.registry.api.RemoteRegistry;
import org.ow2.petals.registry.api.Resource;
import org.ow2.petals.registry.api.config.RemoteConfiguration;
import org.ow2.petals.registry.api.context.Context;
import org.ow2.petals.registry.api.exception.LifeCycleException;
import org.ow2.petals.registry.api.exception.RegistryException;
import org.ow2.petals.registry.api.lifecycle.LifeCycle.STATE;
import org.ow2.petals.registry.api.lifecycle.AbstractLifeCycle;
import org.ow2.petals.registry.api.manager.OutgoingManager;
import org.ow2.petals.registry.api.util.Assert;
import org.ow2.petals.registry.core.factory.RegistryFactory;


/**
 * @author Christophe HAMERLING - eBM WebSourcing
 * @author Charles Casadei - eBM WebSourcing
 */
public class GenericOutgoingManagerImpl implements OutgoingManager {

    private final Log log = LogFactory.getLog(GenericOutgoingManagerImpl.class);
    
    private LocalRegistry localRegistry = null;

    protected AbstractLifeCycle lifeCycle;

    protected Map<String, RemoteRegistry> registries;
    
    protected String name;

    /**
     * Creates a new instance of OutgoingManager
     */
    public GenericOutgoingManagerImpl() {
        this.registries = new HashMap<String, RemoteRegistry>();

        this.lifeCycle = new AbstractLifeCycle() {

            @Override
            protected void doInit(final Context context) throws LifeCycleException {
                //Retrieve the remote registries from the context's topology
                // & add them to the list of the known registries
                List<RemoteConfiguration> remoteRegistries = context.getTopology();
                if (remoteRegistries != null) {
                    for (RemoteConfiguration remoteConfiguration : remoteRegistries) {
                        try {
                            RemoteRegistry remoteRegistry = RegistryFactory.getInstance().loadRemote(remoteConfiguration);
                            if (remoteRegistry != null) {
                                GenericOutgoingManagerImpl.this.putRemote(remoteRegistry);
                            }
                        } catch (RegistryException ex) {
                            //TODO : Add it to a trash list
                            StringBuilder errorSB = new StringBuilder();
                            errorSB.append("Can't initialize the OutgoingManager : ");

                            //Log the error...since this is the only way to have
                            //some feedback on what exception is raised...I like
                            //neanderthalian frameworks that can't properly
                            //handle exceptions...and recompile code in a
                            //way not even god can understand...
                            GenericOutgoingManagerImpl.this.log.error(errorSB.toString(), ex);
                            
                            throw new LifeCycleException(errorSB.toString(), ex);
                        }
                    }

                    Iterator<String> keys = GenericOutgoingManagerImpl.this.registries.keySet().iterator();
                    while (keys.hasNext()) {
                        RemoteRegistry remoteRegistry = GenericOutgoingManagerImpl.this.registries.get(keys.next());
                        try {
                        remoteRegistry.init(context);
                        } catch (LifeCycleException ex) {
                            StringBuilder errorSB = new StringBuilder();
                            
                            errorSB.append("Can't initialize the OutgoingManager : ");

                            GenericOutgoingManagerImpl.this.log.error(errorSB.toString(), ex);

                            throw ex;
                        }
                    }
                }
            }

            @Override
            protected void doStart() throws LifeCycleException {
                Iterator<String> keys = GenericOutgoingManagerImpl.this.registries.keySet().iterator();
                while (keys.hasNext()) {
                    RemoteRegistry remoteRegistry = GenericOutgoingManagerImpl.this.registries.get(keys.next());
                    remoteRegistry.start();
                }
            }

            @Override
            protected void doStop() throws LifeCycleException {
                Iterator<String> keys = GenericOutgoingManagerImpl.this.registries.keySet().iterator();
                Set<RemoteRegistry> failed = new java.util.HashSet<RemoteRegistry>();
                while (keys.hasNext()) {
                    RemoteRegistry remote = GenericOutgoingManagerImpl.this.registries.get(keys.next());
                    try {
                        remote.stop();
                    } catch (LifeCycleException ex) {
                        this.log.warn("Can not stop the remote registry at " + remote.getURI());
                        failed.add(remote);
                    }
                }

                // TODO : Do something with not stopped remote registries...
            }
        };
    }

    /**
     * Get the local registry which is linked to this manager
     *
     * @return the local registry
     *
     * @see org.ow2.petals.registry.api.manager#getLocalRegistry();
     */
    public LocalRegistry getLocalRegistry() {
        return this.localRegistry;
    }

    /**
     * Set the local registry
     *
     * @param localRegistry the local registry
     *
     * @see org.ow2.petals.registry.api.manager#setLocalRegistry(LocalRegistry localRegistry)
     */
    public void setLocalRegistry(final LocalRegistry localRegistry) throws LifeCycleException {
        if (localRegistry == null) {
            throw new LifeCycleException("No local registry found");
        }

        this.localRegistry = localRegistry;
    }

    /**
     * Add a remote registry to the manager. If the remote registry is not
     * initialized nor started, call the good lifecycle methods.
     * 
     * @param
     * @throw exception if the remote registry is not consistent (for example,
     *        no message sender is set)
     *
     * @see org.ow2.petals.registry.api.manager.OutgoingManager#addRemoteRegistry()
     */
    public void addRemoteRegistry(RemoteRegistry remoteRegistry) throws RegistryException {
        if (!this.isStarted()) {
            throw new RegistryException("Can not add a registry in this state " + this.getState());
        }
        if (remoteRegistry == null) {
            throw new RegistryException("Can not add a null remote registry");
        }

        // check if the remote registry is not already in
        if (this.registries.get(remoteRegistry.getName()) != null) {
            this.log.warn("Remote registry " + remoteRegistry.getName() + " is already registered");
            return;
        }

        // call the init on the remote registry if not started???
        if (!remoteRegistry.isStarted()) {
            if (!remoteRegistry.isInitialized()) {
                remoteRegistry.init(this.getLocalRegistry().getContext());
            }
            remoteRegistry.start();
        }
        this.putRemote(remoteRegistry);
    }

    private void putRemote(RemoteRegistry remoteRegistry) {
        this.registries.put(remoteRegistry.getName(), remoteRegistry);
    }

    /**
     *
     * @param lifeCycle
     */
    protected void setLifeCycle(AbstractLifeCycle lifeCycle) {
        this.lifeCycle = lifeCycle;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#init()
     */
    public void init(Context context) throws LifeCycleException {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        this.lifeCycle.init(context);
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#start()
     */
    public void start() throws LifeCycleException {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        this.lifeCycle.start();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#stop()
     */
    public void stop() throws LifeCycleException {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        this.lifeCycle.stop();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#isRunning()
     */
    public boolean isRunning() {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        return this.lifeCycle.isRunning();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#isStarted()
     */
    public boolean isStarted() {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        return this.lifeCycle.isStarted();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#isStopped()
     */
    public boolean isStopped() {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        return this.lifeCycle.isStopped();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#isInitialized()
     */
    public boolean isInitialized() {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        return this.lifeCycle.isInitialized();
    }

    public boolean isNotInitialized() {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        return this.lifeCycle.isNotInitialized();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.registry.core.LifeCycle#getState()
     */
    public STATE getState() {
        Assert.notNull(this.lifeCycle, "No lifecycle");
        return this.lifeCycle.getState();
    }

    /**
     * Put a resource in the registry.
     *
     * @param path
     *            The path to store the resource into. Contains the parent
     *            directory and the resource identifier (ie : in
     *            /path/to/resource/r1 r1 is the resource identifier) if the
     *            resource already exists override, create a new one if not.
     * @param resource
     *            The resource to put
     * 
     * @see org.ow2.petals.registry.api.RemoteOperations#put(String path, Resource resource, Information info)
     */
    public final boolean put(String path, Resource resource, Information information)
            throws RegistryException {
        if (this.log.isDebugEnabled()) {
            this.log.debug("Put the resource " + path);
        }
        return this.doPut(path, resource, information);
    }

   /**
     * Put a resource in the registry.
     *
     * @param path
     *            The path to store the resource into. Contains the parent
     *            directory and the resource identifier (ie : in
     *            /path/to/resource/r1 r1 is the resource identifier) if the
     *            resource already exists override, create a new one if not.
     * @param resource
     *            The resource to put
     *
     */
    protected boolean doPut(String key, Resource resource, Information information)
            throws RegistryException {
        boolean result = false;

        // TODO : store the resource to all the foreign registries in an
        // independant thread (do not wait all...)

        Iterator<String> keys = this.registries.keySet().iterator();
        while (keys.hasNext()) {
            RemoteRegistry registry = this.registries.get(keys.next());
            if (this.log.isDebugEnabled()) {
                this.log.debug("Storing resource to remote registry " + registry);
            }
            try {
                result = registry.put(key, resource, information);
            } catch (RegistryException e) {
                // handle the exception to not penalize other registries
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Can not send put command to registry " + registry.getName());
                    this.log.debug("Remote cause is : " + e.getMessage());
                }
                // TODO : do something with the data which has not been sent...
            }
        }
        // TODO : The result depends on how many remote registries receives the
        // command...
        return result;
    }


    /**
     * Get the element from its path only if the path refers to an element. If
     * the path is a directory, it will returns null, please refer to
     * {@link #getAll(String)}
     *
     * @param path
     * @return the resource if found, otherwise it will throw an exception
     * @throws an
     *             exception there is no resource at the given path or if a
     *             problem occurs.
     *
     * @see org.ow2.petals.registry.api.RemoteOperations#get(String path, Information info)
     */
    public final Resource get(String path, Information information) throws RegistryException {
        if (this.log.isDebugEnabled()) {
            this.log.debug("Get the resource " + path);
        }
        return this.doGet(path, information);
    }

    /**
     * Get the element from its path only if the path refers to an element. If
     * the path is a directory, it will returns null, please refer to
     * {@link #getAll(String)}
     *
     * @param path
     * @return the resource if found, otherwise it will throw an exception
     * @throws an
     *             exception there is no resource at the given path or if a
     *             problem occurs.
     */
    protected Resource doGet(String key, Information information) throws RegistryException {
        Resource resource = null;

        // get the first resource we find in the remote registries
        Iterator<String> keys = this.registries.keySet().iterator();
        boolean found = false;
        while (keys.hasNext() && !found) {
            RemoteRegistry remoteRegistry = this.registries.get(keys.next());
            if (this.log.isDebugEnabled()) {
                this.log.debug("Try to get resource from remote registry " + remoteRegistry);
            }
            try {
                resource = remoteRegistry.get(key, information);
            } catch (RegistryException e) {
                // handle the exception to not penalize other registries
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Can not send get command to registry " + remoteRegistry.getName());
                    this.log.debug("Remote cause is : " + e.getMessage());
                }
            }
            found = (resource != null);
        }

        return resource;
    }

    /**
     * Get all the resources in the given path only if the given path is a
     * directory. If it is, returns all the children, if not, return null... For
     * single resource retrieval, refer to {@link #get(String)}. Be carefull on
     * the implementation side when trying to get all the resources of all the
     * registries. The best way is that the getAll implementation just returns
     * the resources of the local registry ie its vision of the resources of the
     * network at the current time.
     *
     * @param path
     * @return
     * @throws RegistryException
     *             if the path is not a directory path...
     *
     * @see org.ow2.petals.registry.api.RemoteOperations#getAll(String path, Information info)
     */
    public final List<Resource> getAll(String path, Information information)
            throws RegistryException {
        if (this.log.isDebugEnabled()) {
            this.log.debug("Get all the resources from " + path);
        }
        return this.doGetAll(path, information);
    }

    /**
     * Get all the resources in the given path only if the given path is a
     * directory. If it is, returns all the children, if not, return null... For
     * single resource retrieval, refer to {@link #get(String)}. Be carefull on
     * the implementation side when trying to get all the resources of all the
     * registries. The best way is that the getAll implementation just returns
     * the resources of the local registry ie its vision of the resources of the
     * network at the current time.
     *
     * @param path
     * @return
     * @throws RegistryException
     *             if the path is not a directory path...
     */
    protected List<Resource> doGetAll(String key, Information information) throws RegistryException {
        if (this.log.isDebugEnabled()) {
            this.log.debug("Retrieve the ALL the resources under " + key);
        }

        Map<String, Resource> map = new HashMap<String, Resource>();

        Iterator<String> keys = this.registries.keySet().iterator();
        while (keys.hasNext()) {
            RemoteRegistry remoteRegistry = this.registries.get(keys.next());
            if (this.log.isDebugEnabled()) {
                this.log.debug("GetAll to remote registry " + remoteRegistry);
            }

            try {
                List<Resource> partial = remoteRegistry.getAll(key, information);
                // add to map if the resource does not already exists
                if ((partial != null) && (partial.size() != 0)) {
                    for (Resource resource : partial) {
                        if (resource != null) {
                            if (this.log.isDebugEnabled()) {
                                this.log.debug("Find a resource on remote registry " + resource);
                            }

                            if (map.get(resource.getPath()) == null) {
                                this.log.debug("New resource added");
                                map.put(resource.getPath(), resource);
                            } else if (map.get(resource.getPath()).getCreationTime() < resource.getCreationTime()) {

                                if (this.log.isDebugEnabled()) {
                                    this.log.debug("Updating resource which is too old : " + resource);
                                }
                                map.put(resource.getPath(), resource);
                            }
                        }
                    }
                } else {
                    this.log.debug("Can not find anything from remote registry " + remoteRegistry);
                }

            } catch (RegistryException e) {
                // handle the exception to not penalize other registries
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Can not send query command to registry " + remoteRegistry.getName());
                    this.log.debug("Remote cause is : " + e.getMessage());
                }
            }
        }
        return new ArrayList<Resource>(map.values());
    }

    /**
     * Delete the element from the registry. If the path is a directory, delete
     * all the children of the given path (and the path too...)
     *
     * @param key
     * @return true if the resource has been deleted, false if not of if not
     *         found
     *
     * @see org.ow2.petals.registry.api.RemoteOperations#delete(String path, Information information)
     */
    public final boolean delete(String path, Information information) throws RegistryException {
        if (this.log.isDebugEnabled()) {
            this.log.debug("Delete the resource " + path);
        }
        return this.doDelete(path, information);
    }

    /**
     * Delete the element from the registry. If the path is a directory, delete
     * all the children of the given path (and the path too...)
     *
     * @param key
     * @return true if the resource has been deleted, false if not of if not
     *         found
     */
    protected boolean doDelete(String key, Information information) throws RegistryException {
        boolean result = false;

        Iterator<String> keys = this.registries.keySet().iterator();
        while (keys.hasNext()) {
            RemoteRegistry registry = this.registries.get(keys.next());
            if (this.log.isDebugEnabled()) {
                this.log.debug("Remove the resource on " + registry);
            }
            try {
                result = registry.delete(key, information);
            } catch (RegistryException e) {
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Can not send delete command to registry " + registry.getName());
                    this.log.debug("Remote cause is : " + e.getMessage());
                }
            }
        }
        // TODO : The result depends on how many remote registries receives the
        // command...
        return result;
    }

    /**
     *
     * @see org.ow2.petals.registry.api.RemoteOperations#query(Query query, Information info)
     */
    public final List<Resource> query(Query query, Information information)
            throws RegistryException {
        if (this.log.isDebugEnabled()) {
            this.log.debug("Query remote resources with " + query);
        }
        return this.doQuery(query, information);
    }

    /**
     *
     */
    protected List<Resource> doQuery(Query query, Information information) throws RegistryException {
        // FIXME : TODO Define the way to get things... Query all, quey master, query
        // who???????
        // get all the resource from all the foreign registries where the query
        // match
        Iterator<String> keys = this.registries.keySet().iterator();
        Map<String, Resource> map = new HashMap<String, Resource>();
        while (keys.hasNext()) {
            RemoteRegistry remoteRegistry = this.registries.get(keys.next());
            if (this.log.isDebugEnabled()) {
                this.log.debug("Query to remote registry " + remoteRegistry);
            }

            try {
                List<Resource> partial = remoteRegistry.query(query, information);
                // add to map if the resource does not already exists
                if ((partial != null) && (partial.size() != 0)) {
                    for (Resource resource : partial) {
                        if (resource != null) {
                            if (this.log.isDebugEnabled()) {
                                this.log.debug("Find resource on remote registry " + resource);
                            }
                            // will override only if the data is newer than the
                            // current one if any
                            // FIXME : Should have a lastTouchedTime
                            // parameter...
                            if ((map.get(resource.getPath()) != null) && (map.get(resource.getPath()).getCreationTime() < resource.getCreationTime())) {
                                if (this.log.isDebugEnabled()) {
                                    this.log.debug("Putting a new resource in the cache : " + resource);
                                }
                                map.put(resource.getPath(), resource);
                            } else {
                                if (this.log.isDebugEnabled()) {
                                    this.log.debug("Resource is too old and is not kept");
                                }
                            }
                        }
                    }
                } else {
                    this.log.debug("Can not find anything from remote registry " + remoteRegistry);
                }

            } catch (RegistryException e) {
                // handle the exception to not penalize other registries
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Can not send query command to registry " + remoteRegistry.getName());
                    this.log.debug("Remote cause is : " + e.getMessage());
                }
            }
        }
        return new ArrayList<Resource>(map.values());
    }
}
