/*
 * Copyright (c) 2009 EBM Websourcing, http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ow2.petals.registry.core.repository.hibernate;

import java.util.Properties;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.cfg.AnnotationConfiguration;
import org.hibernate.stat.Statistics;
import org.ow2.petals.registry.api.config.DBConfiguration;
import org.ow2.petals.registry.api.exception.RegistryException;
import org.ow2.petals.registry.core.repository.bo.Endpoint;
import org.ow2.petals.registry.core.repository.bo.Interface;
import org.ow2.petals.registry.core.repository.bo.Property;
import org.ow2.petals.registry.core.repository.bo.Resource;

/**
 * @author Christophe HAMERLING - eBM WebSourcing
 * 
 */
public final class HibernateHelper {

    private static SessionFactory sessionFactory = null;

    final static Logger log = Logger.getLogger(HibernateHelper.class);

    /**
     * @return the sessionFactory
     */
    public static SessionFactory getSessionFactory() {
        return sessionFactory;
    }

    /**
     * @return
     * @throws HibernateException
     * @throws DataCollectorConfigException
     */
    public static Session getSession() throws RegistryException {
        if (HibernateHelper.sessionFactory == null) {
            log.debug("hibernate session not configured, shouldn't happen");
            throw new RegistryException(
                    "Can not find the session factory, please configure it first");
        }
        try {
            return HibernateHelper.sessionFactory.openSession();
        } catch (HibernateException e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public static Statistics getSessionStatistics() {
        return HibernateHelper.sessionFactory.getStatistics();
    }
    
    public static void configureSession(DBConfiguration dbConfiguration) {

        try {
            final AnnotationConfiguration configuration = new AnnotationConfiguration();
            final Properties props = new Properties();

            //Used for debugging purpose
            //props.setProperty("show_sql", "false");

            //Configure the DB connection
            props.setProperty("hibernate.connection.driver_class", dbConfiguration.getDriver());
            props.setProperty("hibernate.connection.url", dbConfiguration.getUrl());
            props.setProperty("hibernate.connection.username", dbConfiguration.getUsername());
            props.setProperty("hibernate.connection.password", dbConfiguration.getPassword());
            props.setProperty("hibernate.dialect", dbConfiguration.getDialect());
            
            //Configure session
            props.setProperty("autoReconnect", "true"); 
            props.setProperty("hibernate.connection.autocommit", "true");
            props.setProperty("hibernate.connection.shutdown", "true");
            props.setProperty("hibernate.hbm2ddl.auto", "update");
            props.setProperty("hibernate.query.substitutions", "true");
            

            //C3PO configuration
            if (dbConfiguration.isc3p0Enabled()) {
                props.setProperty("hibernate.connection.provider_class", "org.hibernate.connection.C3P0ConnectionProvider");
                props.setProperty("c3p0.acquire_increment", "1");
                props.setProperty("c3p0.max_size", "2");
                props.setProperty("c3p0.min_size", "1");
                props.setProperty("c3p0.max_statements", "0");
                props.setProperty("c3p0.idle_test_period", "100");
                props.setProperty("c3p0.timeout", "300");
                props.setProperty("c3p0.maxIdleTime", "300");
                props.setProperty("c3p0.maxConnectionAge", "900");
                props.setProperty("c3p0.automaticTestTable", "c3p0_conn_test");
                props.setProperty("c3p0.idleConnectionTestPeriod", "300");
                }
            
            //Configure caching policy
            if (dbConfiguration.isCachingEnabled()) {
                props.setProperty("hibernate.cache.provider_class", "org.hibernate.cache.EhCacheProvider");
                props.setProperty("hibernate.cache.use_structured_cache", "true");
                props.setProperty("hibernate.cache.use_minimal_puts", "true");
                props.setProperty("hibernate.cache.use_query_cache", "true");
                props.setProperty("hibernate.cache.use_second_level_cache", "true");
            }
            else {
                props.setProperty("hibernate.cache.provider_class", "org.hibernate.cache.NoCacheProvider");
            }

            //Register all the above properties in the configuration
            configuration.addProperties(props);

            //Add the entity class and activate the cache
            //nonstrict-read-write is used because write concurrency is unlikely             
            configuration.addAnnotatedClass(Endpoint.class);
            configuration.addAnnotatedClass(Interface.class);
            configuration.addAnnotatedClass(Property.class);
            configuration.addAnnotatedClass(org.ow2.petals.registry.core.repository.bo.Properties.class);

            //Adding cached classes if caching enabled
            if (dbConfiguration.isCachingEnabled()) {
                configuration.setCacheConcurrencyStrategy(Endpoint.class.getCanonicalName(), "nonstrict-read-write");
                configuration.setCacheConcurrencyStrategy(Interface.class.getCanonicalName(), "nonstrict-read-write");
                configuration.setCacheConcurrencyStrategy(Property.class.getCanonicalName(), "nonstrict-read-write");
                configuration.setCacheConcurrencyStrategy(org.ow2.petals.registry.core.repository.bo.Properties.class.getCanonicalName(), "nonstrict-read-write");
            }

            //Initialize the hibernate session
            sessionFactory = configuration.buildSessionFactory();
        } catch (final Throwable e) {
            throw new ExceptionInInitializerError(e);
        }
    }
}
